// Copyright 2024 The Chromium Authors
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

import {SodaEvent} from './soda/types.js';
import {Observer, Unsubscribe} from './utils/observer_list.js';

export enum ModelId {
  // TODO(pihsun): Abstract the "uuid" part of model.
  GEMINI_XXS_IT_BASE = 'ee7c31c2-18e5-405a-b54e-f2607130a15d',
  SUMMARY = '73caa678-45cb-4007-abb9-f04e431376da',
}

export interface Model {
  /**
   * Returns the suggested titles based on content.
   */
  // TODO(pihsun): method to set input options.
  // TODO(pihsun): Other fields needed from mojo, like safety info, ...
  suggestTitles(content: string): Promise<string[]>;

  /**
   * Generates a short summarization of the given content.
   */
  summarize(content: string): Promise<string>;

  /**
   * Closes the model connection.
   *
   * This should release resources used by the model, and no further call of
   * other calls should happen after this.
   */
  close(): void;
}

export interface SodaSession {
  /**
   * Starts the session.
   *
   * Each session can be started at most once.
   */
  start(): Promise<void>;

  /**
   * Adds audio sample to the session.
   *
   * `start` must be called before this.
   *
   * @param samples Array of sample of the audio. Each sample value is in the
   *     range of [-1, 1].
   */
  addAudio(samples: Float32Array): void;

  /**
   * Stops the session and waits for all audio samples to be processed.
   *
   * This can only be called at most once, and `start` must be called before
   * this.
   */
  stop(): Promise<void>;

  /**
   * Add an observer for the result SodaEvent.
   *
   * Since events before subscriptions are not passed to observer on subscribe,
   * this should be called before audio samples are added to ensure that no
   * event is dropped.
   */
  subscribeEvent(observer: Observer<SodaEvent>): Unsubscribe;
}

export abstract class PlatformHandler {
  /**
   * Initializes the platform handler.
   * This should only be called once when the app starts.
   */
  abstract init(): void;

  /**
   * Loads the ML model by the given UUID.
   */
  abstract loadModel(uuid: string): Promise<Model>;

  /**
   * Creates a new soda session for transcription.
   */
  abstract newSodaSession(): Promise<SodaSession>;

  /**
   * Returns the i18n message generated by given `name` and `substitutions`.
   *
   * TODO(pihsun): Change `name` to an enum.
   */
  abstract i18n(name: string, ...substitutions: Array<number|string>): string;

  /**
   * Renders the UI needed on the dev page.
   */
  abstract renderDevUi(): RenderResult;
}
